@@ -14,19 +14,23 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import sys, os
+import typing
+from pathlib import Path
+import sys
 import re
+import argparse
+
 
 class Token:
-    def __init__(self, tid, value):
+    def __init__(self, tid: str, value: str):
         self.tid = tid
         self.value = value
         self.lineno = 0
         self.colno = 0
 
 class Statement:
-    def __init__(self, name, args):
-        self.name = name
+    def __init__(self, name: str, args: list):
+        self.name = name.lower()
         self.args = args
 
 class Lexer:
@@ -43,7 +47,7 @@ class Lexer:
             ('rparen', re.compile(r'\)')),
         ]
 
-    def lex(self, code):
+    def lex(self, code: str) -> typing.Iterator[Token]:
         lineno = 1
         line_start = 0
         loc = 0
@@ -77,13 +81,13 @@ class Lexer:
                     elif tid == 'varexp':
                         yield(Token('varexp', match_text[2:-1]))
                     else:
-                        raise RuntimeError('Wharrgarbl')
+                        raise ValueError('lex: unknown element {}'.format(tid))
                     break
             if not matched:
-                raise RuntimeError('Lexer got confused line %d column %d' % (lineno, col))
+                raise ValueError('Lexer got confused line %d column %d' % (lineno, col))
 
 class Parser:
-    def __init__(self, code):
+    def __init__(self, code: str):
         self.stream = Lexer().lex(code)
         self.getsym()
 
@@ -93,18 +97,18 @@ class Parser:
         except StopIteration:
             self.current = Token('eof', '')
 
-    def accept(self, s):
+    def accept(self, s: str) -> bool:
         if self.current.tid == s:
             self.getsym()
             return True
         return False
 
-    def expect(self, s):
+    def expect(self, s: str) -> bool:
         if self.accept(s):
             return True
-        raise RuntimeError('Expecting %s got %s.' % (s, self.current.tid), self.current.lineno, self.current.colno)
+        raise ValueError('Expecting %s got %s.' % (s, self.current.tid), self.current.lineno, self.current.colno)
 
-    def statement(self):
+    def statement(self) -> Statement:
         cur = self.current
         if self.accept('comment'):
             return Statement('_', [cur.value])
@@ -114,13 +118,16 @@ class Parser:
         self.expect('rparen')
         return Statement(cur.value, args)
 
-    def arguments(self):
+    def arguments(self) -> list:
         args = []
         if self.accept('lparen'):
             args.append(self.arguments())
             self.expect('rparen')
         arg = self.current
-        if self.accept('string') \
+        if self.accept('comment'):
+            rest = self.arguments()
+            args += rest
+        elif self.accept('string') \
                 or self.accept('varexp') \
                 or self.accept('id'):
             args.append(arg)
@@ -128,7 +135,7 @@ class Parser:
             args += rest
         return args
 
-    def parse(self):
+    def parse(self) -> typing.Iterator[Statement]:
         while not self.accept('eof'):
             yield(self.statement())
 
@@ -137,13 +144,13 @@ class Converter:
                      'enable_testing': True,
                      'include': True}
 
-    def __init__(self, cmake_root):
-        self.cmake_root = cmake_root
+    def __init__(self, cmake_root: str):
+        self.cmake_root = Path(cmake_root).expanduser()
         self.indent_unit = '  '
         self.indent_level = 0
-        self.options = []
+        self.options = []  # type: typing.List[tuple]
 
-    def convert_args(self, args, as_array=True):
+    def convert_args(self, args: typing.List[Token], as_array: bool = True) -> str:
         res = []
         if as_array:
             start = '['
@@ -155,19 +162,18 @@ class Converter:
             if i.tid == 'id':
                 res.append("'%s'" % i.value)
             elif i.tid == 'varexp':
-                res.append('%s' % i.value)
+                res.append('%s' % i.value.lower())
             elif i.tid == 'string':
                 res.append("'%s'" % i.value)
             else:
-                print(i)
-                raise RuntimeError('Unknown arg type.')
+                raise ValueError('Unknown arg type {}'.format(i.tid))
         if len(res) > 1:
             return start + ', '.join(res) + end
         if len(res) == 1:
             return res[0]
         return ''
 
-    def write_entry(self, outfile, t):
+    def write_entry(self, outfile: typing.TextIO, t: Statement):
         if t.name in Converter.ignored_funcs:
             return
         preincrement = 0
@@ -197,7 +203,7 @@ class Converter:
                 libcmd = 'static_library'
                 args = [t.args[0]] + t.args[2:]
             else:
-                libcmd = 'static_library'
+                libcmd = 'library'
                 args = t.args
             line = '%s_lib = %s(%s)' % (t.args[0].value, libcmd, self.convert_args(args, False))
         elif t.name == 'add_test':
@@ -220,13 +226,24 @@ class Converter:
                     l = 'cpp'
                 args.append(l)
             args = ["'%s'" % i for i in args]
-            line = 'project(' + ', '.join(args) + ')'
+            line = 'project(' + ', '.join(args) + ", default_options : ['default_library=static'])"
         elif t.name == 'set':
             varname = t.args[0].value.lower()
             line = '%s = %s\n' % (varname, self.convert_args(t.args[1:]))
         elif t.name == 'if':
             postincrement = 1
-            line = 'if %s' % self.convert_args(t.args, False)
+            try:
+                line = 'if %s' % self.convert_args(t.args, False)
+            except AttributeError:  # complex if statements
+                line = t.name
+                for arg in t.args:
+                    if isinstance(arg, Token):
+                        line += ' ' + arg.value
+                    elif isinstance(arg, list):
+                        line += ' ('
+                        for a in arg:
+                            line += ' ' + a.value
+                        line += ' )'
         elif t.name == 'elseif':
             preincrement = -1
             postincrement = 1
@@ -248,32 +265,32 @@ class Converter:
             outfile.write('\n')
         self.indent_level += postincrement
 
-    def convert(self, subdir=''):
-        if subdir == '':
+    def convert(self, subdir: Path = None):
+        if not subdir:
             subdir = self.cmake_root
-        cfile = os.path.join(subdir, 'CMakeLists.txt')
+        cfile = Path(subdir).expanduser() / 'CMakeLists.txt'
         try:
-            with open(cfile) as f:
+            with cfile.open() as f:
                 cmakecode = f.read()
         except FileNotFoundError:
-            print('\nWarning: No CMakeLists.txt in', subdir, '\n')
+            print('\nWarning: No CMakeLists.txt in', subdir, '\n', file=sys.stderr)
             return
         p = Parser(cmakecode)
-        with open(os.path.join(subdir, 'meson.build'), 'w') as outfile:
+        with (subdir / 'meson.build').open('w') as outfile:
             for t in p.parse():
                 if t.name == 'add_subdirectory':
                     # print('\nRecursing to subdir',
-                    #       os.path.join(self.cmake_root, t.args[0].value),
+                    #       self.cmake_root / t.args[0].value,
                     #       '\n')
-                    self.convert(os.path.join(subdir, t.args[0].value))
+                    self.convert(subdir / t.args[0].value)
                     # print('\nReturning to', self.cmake_root, '\n')
                 self.write_entry(outfile, t)
         if subdir == self.cmake_root and len(self.options) > 0:
             self.write_options()
 
     def write_options(self):
-        filename = os.path.join(self.cmake_root, 'meson_options.txt')
-        with open(filename, 'w') as optfile:
+        filename = self.cmake_root / 'meson_options.txt'
+        with filename.open('w') as optfile:
             for o in self.options:
                 (optname, description, default) = o
                 if default is None:
@@ -296,8 +313,8 @@ class Converter:
                 optfile.write(line)
 
 if __name__ == '__main__':
-    if len(sys.argv) != 2:
-        print(sys.argv[0], '<CMake project root>')
-        sys.exit(1)
-    c = Converter(sys.argv[1])
-    c.convert()
+    p = argparse.ArgumentParser(description='Convert CMakeLists.txt to meson.build and meson_options.txt')
+    p.add_argument('cmake_root', help='CMake project root (where top-level CMakeLists.txt is)')
+    P = p.parse_args()
+
+    Converter(P.cmake_root).convert()
